/*
 *      lark-document-loader.c
 *
 *      Copyright 2016-2018 Alex <alex@linuxonly.ru>
 *
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <glib/gstdio.h>

#include <libxml/SAX.h>
#include <libxml/HTMLparser.h>
#include <poppler.h>
#include <gsf/gsf.h>
#include <uchardet/uchardet.h>

#include "lark-document-loader.h"

#define LARK_DOCUMENT_LOADER_ERROR  g_quark_from_static_string("ru.linuxonly.lark.document-loader.error")

struct _LarkDocumentLoader
{
	GObject parent_instance;
	/*Data*/
	gchar *path;
	guint format;
	//gchar *text;
	gchar *title;
	gchar *author;
	gchar *comment;
	/*State*/
	GArray *xmlelements;
	guint xmllevel;
	GString *xmlauthor;
	GString *xmlcomment;
	gchar **epubopfpath;
	GHashTable *epubitems;
	GSList *epubspine;
};

struct _LarkDocumentLoaderClass
{
	GObjectClass parent_class;
};

G_DEFINE_TYPE (LarkDocumentLoader, lark_document_loader, G_TYPE_OBJECT)

enum {
	PROP_0,
	PROP_PATH,
	PROP_FORMAT,
	//PROP_TEXT,
	PROP_TITLE,
	PROP_AUTHOR,
	PROP_COMMENT,
	LAST_PROP
};

static GParamSpec *properties [LAST_PROP];

enum {
	TEXT_CHUNK_READY,
	LAST_SIGNAL
};

static guint signals[LAST_SIGNAL];

enum {
	FB2_ELEMENT_UNKNOWN = 0,
	FB2_ELEMENT_FICTION_BOOK,
	FB2_ELEMENT_PARAGRAPH,
	FB2_ELEMENT_CODE,
	FB2_ELEMENT_EMPTY_LINE,
	FB2_ELEMENT_SECTION,
	FB2_ELEMENT_TITLE,
	FB2_ELEMENT_STRONG,
	FB2_ELEMENT_STRIKETHROUGH,
	FB2_ELEMENT_EMPHASIS,
	FB2_ELEMENT_POEM,
	FB2_ELEMENT_STANZA,
	FB2_ELEMENT_V,
	FB2_ELEMENT_TEXT_DATE,
	FB2_ELEMENT_SUP,
	FB2_ELEMENT_SUB,
	FB2_ELEMENT_SUBTITLE,
	FB2_ELEMENT_TEXT_AUTHOR,
	FB2_ELEMENT_EPIGRAPH,
	FB2_ELEMENT_CITE,
	FB2_ELEMENT_TABLE,
	FB2_ELEMENT_TR,
	FB2_ELEMENT_TH,
	FB2_ELEMENT_TD,
	FB2_ELEMENT_A,
	FB2_ELEMENT_STYLE,
	FB2_ELEMENT_DESCRIPTION,
	FB2_ELEMENT_TITLE_INFO,
	FB2_ELEMENT_GENRE,
	FB2_ELEMENT_AUTHOR,
	FB2_ELEMENT_FIRST_NAME,
	FB2_ELEMENT_LAST_NAME,
	FB2_ELEMENT_BOOK_TITLE,
	FB2_ELEMENT_ANNOTATION,
	FB2_ELEMENT_DATE,
	FB2_ELEMENT_COVER_PAGE,
	FB2_ELEMENT_IMAGE,
	FB2_ELEMENT_LANG,
	FB2_ELEMENT_DOCUMENT_INFO,
	FB2_ELEMENT_PROGRAM_USED,
	FB2_ELEMENT_ID,
	FB2_ELEMENT_VERSION,
	FB2_ELEMENT_HISTORY,
	FB2_ELEMENT_PUBLISH_INFO,
	FB2_ELEMENT_BOOK_NAME,
	FB2_ELEMENT_PUBLISHER,
	FB2_ELEMENT_CITY,
	FB2_ELEMENT_YEAR,
	FB2_ELEMENT_ISBN,
	FB2_ELEMENT_BODY,
	FB2_ELEMENT_BINARY,
	FB2_ELEMENT_OTHER,
};

enum {
	EPUB_CONTAINER_ELEMENT_UNKNOWN = 0,
	EPUB_CONTAINER_ELEMENT_CONTAINER,
	EPUB_CONTAINER_ELEMENT_ROOTFILES,
	EPUB_CONTAINER_ELEMENT_ROOTFILE,
	EPUB_CONTAINER_ELEMENT_OTHER
};

enum {
	EPUB_OPF_ELEMENT_UNKNOWN = 0,
	EPUB_OPF_ELEMENT_PACKAGE,
	EPUB_OPF_ELEMENT_METADATA,
	EPUB_OPF_ELEMENT_PUBLISHER,
	EPUB_OPF_ELEMENT_DESCRIPTION,
	EPUB_OPF_ELEMENT_LANGUAGE,
	EPUB_OPF_ELEMENT_CREATOR,
	EPUB_OPF_ELEMENT_TITLE,
	EPUB_OPF_ELEMENT_DATE,
	EPUB_OPF_ELEMENT_CONTRUBUTOR,
	EPUB_OPF_ELEMENT_IDENTIFER,
	EPUB_OPF_ELEMENT_SUBJECT,
	EPUB_OPF_ELEMENT_META,
	EPUB_OPF_ELEMENT_MANIFEST,
	EPUB_OPF_ELEMENT_ITEM,
	EPUB_OPF_ELEMENT_SPINE,
	EPUB_OPF_ELEMENT_ITEMREF,
	EPUB_OPF_ELEMENT_GUIDE,
	EPUB_OPF_ELEMENT_REFERENCE,
	EPUB_OPF_ELEMENT_OTHER
};

enum {
	EPUB_XHTML_ELEMENT_UNKNOWN = 0,
	EPUB_XHTML_ELEMENT_HTML,
	EPUB_XHTML_ELEMENT_HEAD,
	EPUB_XHTML_ELEMENT_TITLE,
	EPUB_XHTML_ELEMENT_BODY,
	EPUB_XHTML_ELEMENT_PARAGRAPH,
	EPUB_XHTML_ELEMENT_ARTICLE,
	EPUB_XHTML_ELEMENT_DIV,
	EPUB_XHTML_ELEMENT_BLOCKQUOTE,
	EPUB_XHTML_ELEMENT_TABLE,
	EPUB_XHTML_ELEMENT_TR,
	EPUB_XHTML_ELEMENT_TD,
	EPUB_XHTML_ELEMENT_TH,
	EPUB_XHTML_ELEMENT_OL,
	EPUB_XHTML_ELEMENT_UL,
	EPUB_XHTML_ELEMENT_LI,
	EPUB_XHTML_ELEMENT_H1,
	EPUB_XHTML_ELEMENT_H2,
	EPUB_XHTML_ELEMENT_H3,
	EPUB_XHTML_ELEMENT_H4,
	EPUB_XHTML_ELEMENT_H5,
	EPUB_XHTML_ELEMENT_H6,
	EPUB_XHTML_ELEMENT_HR,
	EPUB_XHTML_ELEMENT_BR,
	EPUB_XHTML_ELEMENT_IMG,
	EPUB_XHTML_ELEMENT_A,
	EPUB_XHTML_ELEMENT_EM,
	EPUB_XHTML_ELEMENT_STRONG,
	EPUB_XHTML_ELEMENT_ITALIC,
	EPUB_XHTML_ELEMENT_BOLD,
	EPUB_XHTML_ELEMENT_UNDERLINE,
	EPUB_XHTML_ELEMENT_STRIKETHROUGH,
	EPUB_XHTML_ELEMENT_SUP,
	EPUB_XHTML_ELEMENT_SUB,
	EPUB_XHTML_ELEMENT_SPAN,
	EPUB_XHTML_ELEMENT_OTHER
};

guint lark_document_loader_fb2_element_id_from_name(const gchar *localname)
{
	guint curelement;
	
	g_return_val_if_fail(localname != NULL, FB2_ELEMENT_UNKNOWN);
	
	if (strcmp(localname, "FictionBook") == 0) {
		curelement = FB2_ELEMENT_FICTION_BOOK;
	} else if (strcmp(localname, "p") == 0) {
		curelement = FB2_ELEMENT_PARAGRAPH;
	} else if (strcmp(localname, "code") == 0) {
		curelement = FB2_ELEMENT_CODE;
	} else if (strcmp(localname, "empty-line") == 0) {
		curelement = FB2_ELEMENT_EMPTY_LINE;
	} else if (strcmp(localname, "section") == 0) {
		curelement = FB2_ELEMENT_SECTION;
	} else if (strcmp(localname, "title") == 0) {
		curelement = FB2_ELEMENT_TITLE;
	} else if (strcmp(localname, "strong") == 0) {
		curelement = FB2_ELEMENT_STRONG;
	} else if (strcmp(localname, "emphasis") == 0) {
		curelement = FB2_ELEMENT_EMPHASIS;
	} else if (strcmp(localname, "strikethrough") == 0) {
		curelement = FB2_ELEMENT_STRIKETHROUGH;
	} else if (strcmp(localname, "poem") == 0) {
		curelement = FB2_ELEMENT_POEM;
	} else if (strcmp(localname, "stanza") == 0) {
		curelement = FB2_ELEMENT_STANZA;
	} else if (strcmp(localname, "v") == 0) {
		curelement = FB2_ELEMENT_V;
	} else if (strcmp(localname, "date") == 0) {
		curelement = FB2_ELEMENT_TEXT_DATE;
	} else if (strcmp(localname, "sup") == 0) {
		curelement = FB2_ELEMENT_SUP;
	} else if (strcmp(localname, "sub") == 0) {
		curelement = FB2_ELEMENT_SUB;
	} else if (strcmp(localname, "subtitle") == 0) {
		curelement = FB2_ELEMENT_SUBTITLE;
	} else if (strcmp(localname, "text-author") == 0) {
		curelement = FB2_ELEMENT_TEXT_AUTHOR;
	} else if (strcmp(localname, "epigraph") == 0) {
		curelement = FB2_ELEMENT_EPIGRAPH;
	} else if (strcmp(localname, "cite") == 0) {
		curelement = FB2_ELEMENT_CITE;
	} else if (strcmp(localname, "table") == 0) {
		curelement = FB2_ELEMENT_TABLE;
	} else if (strcmp(localname, "tr") == 0) {
		curelement = FB2_ELEMENT_TR;
	} else if (strcmp(localname, "th") == 0) {
		curelement = FB2_ELEMENT_TH;
	} else if (strcmp(localname, "td") == 0) {
		curelement = FB2_ELEMENT_TD;
	} else if (strcmp(localname, "a") == 0) {
		curelement = FB2_ELEMENT_A;
	} else if (strcmp(localname, "style") == 0) {
		curelement = FB2_ELEMENT_STYLE;
	} else if (strcmp(localname, "description") == 0) {
		curelement = FB2_ELEMENT_DESCRIPTION;
	} else if (strcmp(localname, "title-info") == 0) {
		curelement = FB2_ELEMENT_TITLE_INFO;
	} else if (strcmp(localname, "genre") == 0) {
		curelement = FB2_ELEMENT_GENRE;
	} else if (strcmp(localname, "author") == 0) {
		curelement = FB2_ELEMENT_AUTHOR;
	} else if (strcmp(localname, "first-name") == 0) {
		curelement = FB2_ELEMENT_FIRST_NAME;
	} else if (strcmp(localname, "last-name") == 0) {
		curelement = FB2_ELEMENT_LAST_NAME;
	} else if (strcmp(localname, "book-title") == 0) {
		curelement = FB2_ELEMENT_BOOK_TITLE;
	} else if (strcmp(localname, "annotation") == 0) {
		curelement = FB2_ELEMENT_ANNOTATION;
	} else if (strcmp(localname, "date") == 0) {
		curelement = FB2_ELEMENT_DATE;
	} else if (strcmp(localname, "coverpage") == 0) {
		curelement = FB2_ELEMENT_COVER_PAGE;
	} else if (strcmp(localname, "image") == 0) {
		curelement = FB2_ELEMENT_IMAGE;
	} else if (strcmp(localname, "lang") == 0) {
		curelement = FB2_ELEMENT_LANG;
	} else if (strcmp(localname, "document-info") == 0) {
		curelement = FB2_ELEMENT_DOCUMENT_INFO;
	} else if (strcmp(localname, "program-used") == 0) {
		curelement = FB2_ELEMENT_PROGRAM_USED;
	} else if (strcmp(localname, "id") == 0) {
		curelement = FB2_ELEMENT_ID;
	} else if (strcmp(localname, "version") == 0) {
		curelement = FB2_ELEMENT_VERSION;
	} else if (strcmp(localname, "history") == 0) {
		curelement = FB2_ELEMENT_HISTORY;
	} else if (strcmp(localname, "publish-info") == 0) {
		curelement = FB2_ELEMENT_PUBLISH_INFO;
	} else if (strcmp(localname, "book-name") == 0) {
		curelement = FB2_ELEMENT_BOOK_NAME;
	} else if (strcmp(localname, "publisher") == 0) {
		curelement = FB2_ELEMENT_PUBLISHER;
	} else if (strcmp(localname, "city") == 0) {
		curelement = FB2_ELEMENT_CITY;
	} else if (strcmp(localname, "year") == 0) {
		curelement = FB2_ELEMENT_YEAR;
	} else if (strcmp(localname, "isbn") == 0) {
		curelement = FB2_ELEMENT_ISBN;
	} else if (strcmp(localname, "body") == 0) {
		curelement = FB2_ELEMENT_BODY;
	} else if (strcmp(localname, "binary") == 0) {
		curelement = FB2_ELEMENT_BINARY;
	} else {
		curelement = FB2_ELEMENT_OTHER;
	}
	
	return curelement;
}

guint lark_document_loader_epub_container_element_id_from_name(const gchar *localname)
{
	guint curelement;
	
	g_return_val_if_fail(localname != NULL, EPUB_CONTAINER_ELEMENT_UNKNOWN);
	
	if (strcmp(localname, "container") == 0) {
		curelement = EPUB_CONTAINER_ELEMENT_CONTAINER;
	} else if (strcmp(localname, "rootfiles") == 0) {
		curelement = EPUB_CONTAINER_ELEMENT_ROOTFILES;
	} else if (strcmp(localname, "rootfile") == 0) {
		curelement = EPUB_CONTAINER_ELEMENT_ROOTFILE;
	} else {
		curelement = EPUB_CONTAINER_ELEMENT_OTHER;
	}
	
	return curelement;
}

guint lark_document_loader_epub_opf_element_id_from_name(const gchar *localname)
{
	guint curelement;
	
	g_return_val_if_fail(localname != NULL, EPUB_OPF_ELEMENT_UNKNOWN);
	
	if (strcmp(localname, "package") == 0) {
		curelement = EPUB_OPF_ELEMENT_PACKAGE;
	} else if (strcmp(localname, "metadata") == 0) {
		curelement = EPUB_OPF_ELEMENT_METADATA;
	} else if (strcmp(localname, "publisher") == 0) {
		curelement = EPUB_OPF_ELEMENT_PUBLISHER;
	} else if (strcmp(localname, "description") == 0) {
		curelement = EPUB_OPF_ELEMENT_DESCRIPTION;
	} else if (strcmp(localname, "language") == 0) {
		curelement = EPUB_OPF_ELEMENT_LANGUAGE;
	} else if (strcmp(localname, "creator") == 0) {
		curelement = EPUB_OPF_ELEMENT_CREATOR;
	} else if (strcmp(localname, "title") == 0) {
		curelement = EPUB_OPF_ELEMENT_TITLE;
	} else if (strcmp(localname, "date") == 0) {
		curelement = EPUB_OPF_ELEMENT_DATE;
	} else if (strcmp(localname, "contributor") == 0) {
		curelement = EPUB_OPF_ELEMENT_CONTRUBUTOR;
	} else if (strcmp(localname, "identifier") == 0) {
		curelement = EPUB_OPF_ELEMENT_IDENTIFER;
	} else if (strcmp(localname, "subject") == 0) {
		curelement = EPUB_OPF_ELEMENT_SUBJECT;
	} else if (strcmp(localname, "meta") == 0) {
		curelement = EPUB_OPF_ELEMENT_META;
	} else if (strcmp(localname, "manifest") == 0) {
		curelement = EPUB_OPF_ELEMENT_MANIFEST;
	} else if (strcmp(localname, "item") == 0) {
		curelement = EPUB_OPF_ELEMENT_ITEM;
	} else if (strcmp(localname, "spine") == 0) {
		curelement = EPUB_OPF_ELEMENT_SPINE;
	} else if (strcmp(localname, "itemref") == 0) {
		curelement = EPUB_OPF_ELEMENT_ITEMREF;
	} else if (strcmp(localname, "guide") == 0) {
		curelement = EPUB_OPF_ELEMENT_GUIDE;
	} else if (strcmp(localname, "reference") == 0) {
		curelement = EPUB_OPF_ELEMENT_REFERENCE;
	} else {
		curelement = EPUB_OPF_ELEMENT_OTHER;
	}
	
	return curelement;
}

guint lark_document_loader_epub_xhtml_element_id_from_name(const gchar *localname)
{
	guint curelement;
	
	g_return_val_if_fail(localname != NULL, EPUB_XHTML_ELEMENT_UNKNOWN);
	
	if (strcmp(localname, "html") == 0) {
		curelement = EPUB_XHTML_ELEMENT_HTML;
	} else if (strcmp(localname, "head") == 0) {
		curelement = EPUB_XHTML_ELEMENT_HEAD;
	} else if (strcmp(localname, "title") == 0) {
		curelement = EPUB_XHTML_ELEMENT_TITLE;
	} else if (strcmp(localname, "body") == 0) {
		curelement = EPUB_XHTML_ELEMENT_BODY;
	} else if (strcmp(localname, "p") == 0) {
		curelement = EPUB_XHTML_ELEMENT_PARAGRAPH;
	} else if (strcmp(localname, "article") == 0) {
		curelement = EPUB_XHTML_ELEMENT_ARTICLE;
	} else if (strcmp(localname, "div") == 0) {
		curelement = EPUB_XHTML_ELEMENT_DIV;
	} else if (strcmp(localname, "blockquote") == 0) {
		curelement = EPUB_XHTML_ELEMENT_BLOCKQUOTE;
	} else if (strcmp(localname, "table") == 0) {
		curelement = EPUB_XHTML_ELEMENT_TABLE;
	} else if (strcmp(localname, "tr") == 0) {
		curelement = EPUB_XHTML_ELEMENT_TR;
	} else if (strcmp(localname, "td") == 0) {
		curelement = EPUB_XHTML_ELEMENT_TD;
	} else if (strcmp(localname, "th") == 0) {
		curelement = EPUB_XHTML_ELEMENT_TH;
	} else if (strcmp(localname, "ol") == 0) {
		curelement = EPUB_XHTML_ELEMENT_OL;
	} else if (strcmp(localname, "ul") == 0) {
		curelement = EPUB_XHTML_ELEMENT_UL;
	} else if (strcmp(localname, "li") == 0) {
		curelement = EPUB_XHTML_ELEMENT_LI;
	} else if (strcmp(localname, "h1") == 0) {
		curelement = EPUB_XHTML_ELEMENT_H1;
	} else if (strcmp(localname, "h2") == 0) {
		curelement = EPUB_XHTML_ELEMENT_H2;
	} else if (strcmp(localname, "h3") == 0) {
		curelement = EPUB_XHTML_ELEMENT_H3;
	} else if (strcmp(localname, "h4") == 0) {
		curelement = EPUB_XHTML_ELEMENT_H4;
	} else if (strcmp(localname, "h5") == 0) {
		curelement = EPUB_XHTML_ELEMENT_H5;
	} else if (strcmp(localname, "h6") == 0) {
		curelement = EPUB_XHTML_ELEMENT_H6;
	} else if (strcmp(localname, "hr") == 0) {
		curelement = EPUB_XHTML_ELEMENT_HR;
	} else if (strcmp(localname, "br") == 0) {
		curelement = EPUB_XHTML_ELEMENT_BR;
	} else if (strcmp(localname, "img") == 0) {
		curelement = EPUB_XHTML_ELEMENT_IMG;
	} else if (strcmp(localname, "a") == 0) {
		curelement = EPUB_XHTML_ELEMENT_A;
	} else if (strcmp(localname, "em") == 0) {
		curelement = EPUB_XHTML_ELEMENT_EM;
	} else if (strcmp(localname, "strong") == 0) {
		curelement = EPUB_XHTML_ELEMENT_STRONG;
	} else if (strcmp(localname, "i") == 0) {
		curelement = EPUB_XHTML_ELEMENT_ITALIC;
	} else if (strcmp(localname, "b") == 0) {
		curelement = EPUB_XHTML_ELEMENT_BOLD;
	} else if (strcmp(localname, "u") == 0) {
		curelement = EPUB_XHTML_ELEMENT_UNDERLINE;
	} else if (strcmp(localname, "s") == 0) {
		curelement = EPUB_XHTML_ELEMENT_STRIKETHROUGH;
	} else if (strcmp(localname, "sup") == 0) {
		curelement = EPUB_XHTML_ELEMENT_SUP;
	} else if (strcmp(localname, "sub") == 0) {
		curelement = EPUB_XHTML_ELEMENT_SUB;
	} else if (strcmp(localname, "span") == 0) {
		curelement = EPUB_XHTML_ELEMENT_SPAN;
	} else {
		curelement = EPUB_XHTML_ELEMENT_OTHER;
	}
	
	return curelement;
}

gint lark_document_loader_fb2_get_element_id_at_level(LarkDocumentLoader *self, guint level)
{
	g_return_val_if_fail(self != NULL, FB2_ELEMENT_UNKNOWN);
	g_return_val_if_fail(level <= self->xmlelements->len, FB2_ELEMENT_UNKNOWN);
	
	return g_array_index(self->xmlelements, guint, level);
}

gboolean lark_document_loader_fb2_set_element_id_at_level(LarkDocumentLoader *self, guint level, guint id)
{
	guint *element;
	
	g_return_val_if_fail(self != NULL, FALSE);
	g_return_val_if_fail(id <= FB2_ELEMENT_OTHER, FALSE);
	
	if (self->xmlelements->len < level + 1) {
		g_array_append_val(self->xmlelements, id);
	} else {
		element = &g_array_index(self->xmlelements, guint, level);
		*element = id;
	}
	
	return TRUE;
}

gboolean lark_document_loader_fb2_append_element_id(LarkDocumentLoader *self, guint id)
{
	g_return_val_if_fail(self != NULL, FALSE);
	g_return_val_if_fail(id <= FB2_ELEMENT_OTHER, FALSE);
	
	g_array_append_val(self->xmlelements, id);
	
	return TRUE;
}

static void lark_document_loader_fb2_on_start_element(gpointer data, const xmlChar *name, const xmlChar **attrs)
{
	LarkDocumentLoader *self = (LarkDocumentLoader *)data;
	guint element;
	
	g_return_if_fail(self != NULL);
	
	element = lark_document_loader_fb2_element_id_from_name((const gchar *)name);
	
	lark_document_loader_fb2_set_element_id_at_level(self, self->xmllevel, element);
	
	self->xmllevel++;
}

static void lark_document_loader_fb2_on_end_element(gpointer data, const xmlChar *name)
{
	LarkDocumentLoader *self = (LarkDocumentLoader *)data;
	guint element, stelement;
	
	g_return_if_fail(self != NULL);
	
	element = lark_document_loader_fb2_element_id_from_name((const gchar *)name);
	stelement = lark_document_loader_fb2_get_element_id_at_level(self, self->xmllevel-1);
	
	if (stelement == element) {
		if (element == FB2_ELEMENT_PARAGRAPH) {
			/*Paragraph*/
			if (lark_document_loader_fb2_get_element_id_at_level(self, 1) == FB2_ELEMENT_DESCRIPTION) {
				/*Title info*/
				if (lark_document_loader_fb2_get_element_id_at_level(self, 2) == FB2_ELEMENT_TITLE_INFO) {
					/*Paragraph of annotation*/
					if (lark_document_loader_fb2_get_element_id_at_level(self, 3) == FB2_ELEMENT_ANNOTATION) {
						g_string_append_c(self->xmlcomment, ' ');
					}
				}
			} else if (lark_document_loader_fb2_get_element_id_at_level(self, 1) == FB2_ELEMENT_BODY) {
				/*Paragraph of book text*/
				g_signal_emit_by_name(G_OBJECT(self), "text-chunk-ready", "\n", 1);
			}
		} else if (element == FB2_ELEMENT_TEXT_AUTHOR) {
			if (lark_document_loader_fb2_get_element_id_at_level(self, 1) == FB2_ELEMENT_BODY) {
				/*Text author line*/
				g_signal_emit_by_name(G_OBJECT(self), "text-chunk-ready", "\n", 1);
			}
		} else if (element == FB2_ELEMENT_TEXT_DATE) {
			if (lark_document_loader_fb2_get_element_id_at_level(self, 1) == FB2_ELEMENT_BODY) {
				/*Text date*/
				g_signal_emit_by_name(G_OBJECT(self), "text-chunk-ready", "\n", 1);
			}
		} else if (element == FB2_ELEMENT_V) {
			if (lark_document_loader_fb2_get_element_id_at_level(self, 1) == FB2_ELEMENT_BODY) {
				/*Poem string*/
				g_signal_emit_by_name(G_OBJECT(self), "text-chunk-ready", "\n", 1);
			}
		} else if ((stelement == FB2_ELEMENT_TH) || (stelement == FB2_ELEMENT_TD)) {
			if (lark_document_loader_fb2_get_element_id_at_level(self, 1) == FB2_ELEMENT_BODY) {
				/*Table cell value*/
				g_signal_emit_by_name(G_OBJECT(self), "text-chunk-ready", "\n", 1);
			}
		} else if (element == FB2_ELEMENT_EMPTY_LINE) {
			/*Just empty line, nothing else*/
			g_signal_emit_by_name(G_OBJECT(self), "text-chunk-ready", "\n", 1);
		}
		
		lark_document_loader_fb2_set_element_id_at_level(self, self->xmllevel-1, FB2_ELEMENT_UNKNOWN);
		self->xmllevel--;
	}
}

static void lark_document_loader_fb2_on_characters(gpointer data, const xmlChar *ch, gint len)
{
	LarkDocumentLoader *self = (LarkDocumentLoader *)data;
	gchar chars[len + 1];
    guint stelement, prelement;
    
    g_return_if_fail(self != NULL);
    
    strncpy(chars, (const gchar *)ch, len);
    chars[len] = '\0';
    
    stelement = lark_document_loader_fb2_get_element_id_at_level(self, self->xmllevel-1);
	
	if (lark_document_loader_fb2_get_element_id_at_level(self, 1) == FB2_ELEMENT_DESCRIPTION) {
		/*Description section*/
		if (lark_document_loader_fb2_get_element_id_at_level(self, 2) == FB2_ELEMENT_TITLE_INFO) {
			/*Book information*/
			if (lark_document_loader_fb2_get_element_id_at_level(self, 3) == FB2_ELEMENT_BOOK_TITLE) {
				/*Title*/
				self->title = g_strdup(chars);
			} else if (lark_document_loader_fb2_get_element_id_at_level(self, 3) == FB2_ELEMENT_AUTHOR) {
				/*Author name*/
				if (lark_document_loader_fb2_get_element_id_at_level(self, 4) == FB2_ELEMENT_FIRST_NAME) {
					if (self->xmlauthor->len == 0) {
						g_string_assign(self->xmlauthor, chars);
					} else {
						g_string_append_printf(self->xmlauthor, " %s", chars);
					}
				} else if (lark_document_loader_fb2_get_element_id_at_level(self, 4) == FB2_ELEMENT_LAST_NAME) {
					if (self->xmlauthor->len == 0) {
						g_string_assign(self->xmlauthor, chars);
					} else {
						g_string_append_printf(self->xmlauthor, " %s", chars);
					}
				}
			} else if (lark_document_loader_fb2_get_element_id_at_level(self, 3) == FB2_ELEMENT_ANNOTATION) {
				/*Annotation*/
				if (lark_document_loader_fb2_get_element_id_at_level(self, 4) == FB2_ELEMENT_PARAGRAPH) {
					if (self->xmlcomment->len == 0) {
						g_string_assign(self->xmlcomment, chars);
					} else {
						g_string_append(self->xmlcomment, chars);
					}				
				}
			}
		}
	} else if (lark_document_loader_fb2_get_element_id_at_level(self, 1) == FB2_ELEMENT_BODY) {
		/*Text section*/
		if (stelement == FB2_ELEMENT_PARAGRAPH) {
			/*Text paragraph*/
			g_signal_emit_by_name(G_OBJECT(self), "text-chunk-ready", chars, len);
		} else if (stelement == FB2_ELEMENT_CODE) {
			prelement = lark_document_loader_fb2_get_element_id_at_level(self, self->xmllevel-2);
			if ((prelement == FB2_ELEMENT_PARAGRAPH) || (prelement == FB2_ELEMENT_V) || (prelement == FB2_ELEMENT_SUBTITLE) || (prelement == FB2_ELEMENT_TEXT_AUTHOR)) {
				/*Program code*/
				g_signal_emit_by_name(G_OBJECT(self), "text-chunk-ready", chars, len);
			}
		} else if (stelement == FB2_ELEMENT_STRONG) {
			prelement = lark_document_loader_fb2_get_element_id_at_level(self, self->xmllevel-2);
			if ((prelement == FB2_ELEMENT_PARAGRAPH) || (prelement == FB2_ELEMENT_V) || (prelement == FB2_ELEMENT_SUBTITLE) || (prelement == FB2_ELEMENT_TEXT_AUTHOR)) {
				/*Bold text paragraph*/
				g_signal_emit_by_name(G_OBJECT(self), "text-chunk-ready", chars, len);
			}
		} else if (stelement == FB2_ELEMENT_EMPHASIS) {
			prelement = lark_document_loader_fb2_get_element_id_at_level(self, self->xmllevel-2);
			if ((prelement == FB2_ELEMENT_PARAGRAPH) || (prelement == FB2_ELEMENT_V) || (prelement == FB2_ELEMENT_SUBTITLE) || (prelement == FB2_ELEMENT_TEXT_AUTHOR)) {
				/*Italic text paragraph*/
				g_signal_emit_by_name(G_OBJECT(self), "text-chunk-ready", chars, len);
			}
		} else if (stelement == FB2_ELEMENT_STRIKETHROUGH) {
			prelement = lark_document_loader_fb2_get_element_id_at_level(self, self->xmllevel-2);
			if ((prelement == FB2_ELEMENT_PARAGRAPH) || (prelement == FB2_ELEMENT_V) || (prelement == FB2_ELEMENT_SUBTITLE) || (prelement == FB2_ELEMENT_TEXT_AUTHOR)) {
				/*Strikethrough text paragraph*/
				g_signal_emit_by_name(G_OBJECT(self), "text-chunk-ready", chars, len);
			}
		} else if (stelement == FB2_ELEMENT_SUP) {
			prelement = lark_document_loader_fb2_get_element_id_at_level(self, self->xmllevel-2);
			if ((prelement == FB2_ELEMENT_PARAGRAPH) || (prelement == FB2_ELEMENT_V) || (prelement == FB2_ELEMENT_SUBTITLE) || (prelement == FB2_ELEMENT_TEXT_AUTHOR)) {
				/*Superscript text*/
				g_signal_emit_by_name(G_OBJECT(self), "text-chunk-ready", chars, len);
			}
		} else if (stelement == FB2_ELEMENT_SUB) {
			prelement = lark_document_loader_fb2_get_element_id_at_level(self, self->xmllevel-2);
			if ((prelement == FB2_ELEMENT_PARAGRAPH) || (prelement == FB2_ELEMENT_V) || (prelement == FB2_ELEMENT_SUBTITLE) || (prelement == FB2_ELEMENT_TEXT_AUTHOR)) {
				/*Subscript text*/
				g_signal_emit_by_name(G_OBJECT(self), "text-chunk-ready", chars, len);
			}
		} else if (stelement == FB2_ELEMENT_V) {
			prelement = lark_document_loader_fb2_get_element_id_at_level(self, self->xmllevel-2);
			if (prelement == FB2_ELEMENT_STANZA) {
				/*Part of poem*/
				g_signal_emit_by_name(G_OBJECT(self), "text-chunk-ready", chars, len);
			}
		} else if (stelement == FB2_ELEMENT_A) {
			prelement = lark_document_loader_fb2_get_element_id_at_level(self, self->xmllevel-2);
			if (prelement == FB2_ELEMENT_PARAGRAPH) {
				/*Link name*/
				g_signal_emit_by_name(G_OBJECT(self), "text-chunk-ready", chars, len);
			}
		} else if (stelement == FB2_ELEMENT_STYLE) {
			prelement = lark_document_loader_fb2_get_element_id_at_level(self, self->xmllevel-2);
			if (prelement == FB2_ELEMENT_PARAGRAPH) {
				/*Link name*/
				g_signal_emit_by_name(G_OBJECT(self), "text-chunk-ready", chars, len);
			}
		} else if ((stelement == FB2_ELEMENT_TH) || (stelement == FB2_ELEMENT_TD)) {
			prelement = lark_document_loader_fb2_get_element_id_at_level(self, self->xmllevel-2);
			if (prelement == FB2_ELEMENT_TR) {
				/*Table cell value*/
				g_signal_emit_by_name(G_OBJECT(self), "text-chunk-ready", chars, len);
			}
		} else if (stelement == FB2_ELEMENT_TEXT_AUTHOR) {
			prelement = lark_document_loader_fb2_get_element_id_at_level(self, self->xmllevel-2);
			if ((prelement == FB2_ELEMENT_POEM) || (prelement == FB2_ELEMENT_CITE) || (prelement == FB2_ELEMENT_EPIGRAPH)) {
				/*Authors name*/
				g_signal_emit_by_name(G_OBJECT(self), "text-chunk-ready", chars, len);
			}
		} else if (stelement == FB2_ELEMENT_TEXT_DATE) {
			prelement = lark_document_loader_fb2_get_element_id_at_level(self, self->xmllevel-2);
			if (prelement == FB2_ELEMENT_POEM) {
				/*Poem date name*/
				g_signal_emit_by_name(G_OBJECT(self), "text-chunk-ready", chars, len);
			}
		}
	} else if (lark_document_loader_fb2_get_element_id_at_level(self, 1) == FB2_ELEMENT_BINARY) {
		/*Binary data section*/
	}
}

static void lark_document_loader_epub_container_on_start_element(gpointer data, const xmlChar *name, const xmlChar *prefix, const xmlChar *URI, int nb_namespaces, const xmlChar **namespaces, int nb_attributes, int nb_defaulted, const xmlChar **attributes)
{
	LarkDocumentLoader *self = (LarkDocumentLoader *)data;
	guint element;
	xmlChar *attrname, *attrprefix __attribute__((unused)), *attrURI __attribute__((unused)), *attrvalstart, *attrvalend;
	gint a;
	gsize valuesize;
	gchar value[256];
	gchar *fullpath, *mediatype;
	
	g_return_if_fail(self != NULL);
	
	element = lark_document_loader_epub_container_element_id_from_name((const gchar *)name);
	
	if (self->epubopfpath == NULL) {
		if (element == EPUB_CONTAINER_ELEMENT_ROOTFILE) {
			/*Prepare storage*/
			fullpath = NULL;
			mediatype = NULL;
			/*Enemerate attributes*/
			for (a = 0; a < nb_attributes; a++) {
				attrname = (xmlChar *)attributes[a * 5 + 0];
				attrprefix = (xmlChar *)attributes[a * 5 + 1];
				attrURI = (xmlChar *)attributes[a * 5 + 2];
				attrvalstart = (xmlChar *)attributes[a * 5 + 3];
				attrvalend = (xmlChar *)attributes[a * 5 + 4];
				valuesize = attrvalend - attrvalstart;
				if ((attrname != NULL) && (valuesize > 0)) {
					memset(value, 0, sizeof(value));
					strncpy(value, (const gchar *)attrvalstart, valuesize);
					if (g_str_equal(attrname, "full-path")) {
						fullpath = g_strdup(value);
					} else if (g_str_equal(attrname, "media-type")) {
						mediatype = g_strdup(value);
					}
				}
			}
			/*Save first OPF file path*/
			if ((fullpath != NULL) && (mediatype != NULL)) {
				if (g_str_equal(mediatype, "application/oebps-package+xml")) {
					self->epubopfpath = g_strsplit(fullpath, "/", -1);
				}
			}
			/*Free resources*/
			if (fullpath != NULL) {
				g_free(fullpath);
			}
			if (mediatype != NULL) {
				g_free(mediatype);
			}
		}
	}
}

static void lark_document_loader_epub_opf_on_start_element(gpointer data, const xmlChar *name, const xmlChar *prefix, const xmlChar *URI, int nb_namespaces, const xmlChar **namespaces, int nb_attributes, int nb_defaulted, const xmlChar **attributes)
{
	LarkDocumentLoader *self = (LarkDocumentLoader *)data;
	guint element;
	xmlChar *attrname, *attrprefix __attribute__((unused)), *attrURI __attribute__((unused)), *attrvalstart, *attrvalend;
	gint a;
	gsize valuesize;
	gchar value[256];
	gchar *itemhref, *itemid, *itemtype, *idref;
	
	g_return_if_fail(self != NULL);
	
	element = lark_document_loader_epub_opf_element_id_from_name((const gchar *)name);
	
	lark_document_loader_fb2_set_element_id_at_level(self, self->xmllevel, element);
	
	if ((element == EPUB_OPF_ELEMENT_ITEM) || (element == EPUB_OPF_ELEMENT_ITEMREF)) {
		/*Prepare storage*/
		itemhref = NULL;
		itemid = NULL;
		itemtype = NULL;
		idref = NULL;
		/*Enemerate attributes*/
		for (a = 0; a < nb_attributes; a++) {
			attrname = (xmlChar *)attributes[a * 5 + 0];
			attrprefix = (xmlChar *)attributes[a * 5 + 1];
			attrURI = (xmlChar *)attributes[a * 5 + 2];
			attrvalstart = (xmlChar *)attributes[a * 5 + 3];
			attrvalend = (xmlChar *)attributes[a * 5 + 4];
			valuesize = attrvalend - attrvalstart;
			if ((attrname != NULL) && (valuesize > 0)) {
				memset(value, 0, sizeof(value));
				strncpy(value, (const gchar *)attrvalstart, valuesize);
				if (element == EPUB_OPF_ELEMENT_ITEM) {
					if (g_str_equal(attrname, "href")) {
						itemhref = g_strdup(value);
					} else if (g_str_equal(attrname, "id")) {
						itemid = g_strdup(value);
					} else if (g_str_equal(attrname, "media-type")) {
						itemtype = g_strdup(value);
					}
				} else if (element == EPUB_OPF_ELEMENT_ITEMREF) {
					if (g_str_equal(attrname, "idref")) {
						idref = g_strdup(value);
					}
				}
			}
		}
		/*Save valuable information*/
		if (element == EPUB_OPF_ELEMENT_ITEM) {
			if ((itemhref != NULL) && (itemid != NULL) && (itemtype != NULL)) {
				if ((g_str_equal(itemtype, "application/xhtml+xml")) || (g_str_equal(itemtype, "text/html"))) {
					g_hash_table_insert(self->epubitems, itemid, itemhref);
				} else {
					g_free(itemhref);
					g_free(itemid);
				}
				g_free(itemtype);
			} else {
				if (itemhref != NULL) {
					g_free(itemhref);
				}
				if (itemid != NULL) {
					g_free(itemid);
				}
				if (itemtype != NULL) {
					g_free(itemtype);
				}
			}
		} else if (element == EPUB_OPF_ELEMENT_ITEMREF) {
			if (idref != NULL) {
				itemhref = g_hash_table_lookup(self->epubitems, idref);
				if (itemhref != NULL) {
					self->epubspine = g_slist_prepend(self->epubspine, itemhref);
				}
				g_free(idref);
			}
		}
	}
	
	self->xmllevel++;
}

static void lark_document_loader_epub_opf_on_end_element(gpointer data, const xmlChar *name, const xmlChar *prefix, const xmlChar *URI)
{
	LarkDocumentLoader *self = (LarkDocumentLoader *)data;
	guint element, stelement;
	
	g_return_if_fail(self != NULL);
	
	element = lark_document_loader_epub_opf_element_id_from_name((const gchar *)name);
	stelement = lark_document_loader_fb2_get_element_id_at_level(self, self->xmllevel-1);
	
	if (stelement == element) {
		lark_document_loader_fb2_set_element_id_at_level(self, self->xmllevel-1, EPUB_OPF_ELEMENT_UNKNOWN);
		self->xmllevel--;
	}
}

static void lark_document_loader_epub_opf_on_characters(gpointer data, const xmlChar *ch, gint len)
{
	LarkDocumentLoader *self = (LarkDocumentLoader *)data;
	gchar chars[len + 1];
	
	g_return_if_fail(self != NULL);
	
	strncpy(chars, (const gchar *)ch, len);
	chars[len] = '\0';
	
	if (lark_document_loader_fb2_get_element_id_at_level(self, 1) == EPUB_OPF_ELEMENT_METADATA) {
		/*Metadata section*/
		if (lark_document_loader_fb2_get_element_id_at_level(self, 2) == EPUB_OPF_ELEMENT_TITLE) {
			/*Title*/
			if (chars[0] != '\n') {
				self->title = g_strdup(chars);
			}
		} else if (lark_document_loader_fb2_get_element_id_at_level(self, 2) == EPUB_OPF_ELEMENT_CREATOR) {
			/*Author name*/
			if (chars[0] != '\n') {
				g_string_assign(self->xmlauthor, chars);
			}
		} else if (lark_document_loader_fb2_get_element_id_at_level(self, 2) == EPUB_OPF_ELEMENT_DESCRIPTION) {
			/*Annotation*/
			if (chars[0] != '\n') {
				if (self->xmlcomment->len == 0) {
					g_string_assign(self->xmlcomment, chars);
				} else {
					g_string_append(self->xmlcomment, chars);
				}
			}
		}
	}
}

static void lark_document_loader_epub_xhtml_on_start_element(gpointer data, const xmlChar *name, const xmlChar **atts)
{
	LarkDocumentLoader *self = (LarkDocumentLoader *)data;
	guint element;
		
	g_return_if_fail(self != NULL);
	
	element = lark_document_loader_epub_xhtml_element_id_from_name((const gchar *)name);
	
	lark_document_loader_fb2_set_element_id_at_level(self, self->xmllevel, element);
	
	self->xmllevel++;
}

static void lark_document_loader_epub_xhtml_on_end_element(gpointer data, const xmlChar *name)
{
	LarkDocumentLoader *self = (LarkDocumentLoader *)data;
	guint element, stelement;
	
	g_return_if_fail(self != NULL);
	
	element = lark_document_loader_epub_xhtml_element_id_from_name((const gchar *)name);
	stelement = lark_document_loader_fb2_get_element_id_at_level(self, self->xmllevel-1);
	
	if (stelement == element) {
		if ((element == EPUB_XHTML_ELEMENT_PARAGRAPH) || (element == EPUB_XHTML_ELEMENT_ARTICLE) || (element == EPUB_XHTML_ELEMENT_DIV) || (element == EPUB_XHTML_ELEMENT_BLOCKQUOTE)) {
			if (lark_document_loader_fb2_get_element_id_at_level(self, 1) == EPUB_XHTML_ELEMENT_BODY) {
				/*Paragraph of book text*/
				g_signal_emit_by_name(G_OBJECT(self), "text-chunk-ready", "\n", 1);
			}
		} else if ((stelement == EPUB_XHTML_ELEMENT_TH) || (stelement == EPUB_XHTML_ELEMENT_TR)) {
			if (lark_document_loader_fb2_get_element_id_at_level(self, 1) == EPUB_XHTML_ELEMENT_BODY) {
				/*Table row*/
				g_signal_emit_by_name(G_OBJECT(self), "text-chunk-ready", "\n", 1);
			}
		} else if (stelement == EPUB_XHTML_ELEMENT_TD) {
			if (lark_document_loader_fb2_get_element_id_at_level(self, 1) == EPUB_XHTML_ELEMENT_BODY) {
				/*Table cell*/
				g_signal_emit_by_name(G_OBJECT(self), "text-chunk-ready", " ", 1);
			}
		} else if (stelement == EPUB_XHTML_ELEMENT_LI) {
			if (lark_document_loader_fb2_get_element_id_at_level(self, 1) == EPUB_XHTML_ELEMENT_BODY) {
				/*List item*/
				g_signal_emit_by_name(G_OBJECT(self), "text-chunk-ready", "\n", 1);
			}
		} else if ((stelement == EPUB_XHTML_ELEMENT_H1) || (stelement == EPUB_XHTML_ELEMENT_H2) || (stelement == EPUB_XHTML_ELEMENT_H3) || (stelement == EPUB_XHTML_ELEMENT_H4) || (stelement == EPUB_XHTML_ELEMENT_H5) || (stelement == EPUB_XHTML_ELEMENT_H6)) {
			if (lark_document_loader_fb2_get_element_id_at_level(self, 1) == EPUB_XHTML_ELEMENT_BODY) {
				/*Header*/
				g_signal_emit_by_name(G_OBJECT(self), "text-chunk-ready", "\n", 1);
			}	
		} else if (stelement == EPUB_XHTML_ELEMENT_IMG) {
			if (lark_document_loader_fb2_get_element_id_at_level(self, 1) == EPUB_XHTML_ELEMENT_BODY) {
				/*Image*/
				g_signal_emit_by_name(G_OBJECT(self), "text-chunk-ready", "\n", 1);
			}
		} else if (stelement == EPUB_XHTML_ELEMENT_A) {
			if (lark_document_loader_fb2_get_element_id_at_level(self, 1) == EPUB_XHTML_ELEMENT_BODY) {
				/*Link*/
				/*g_signal_emit_by_name(G_OBJECT(self), "text-chunk-ready", "\n", 1);*/
			}
		} else if ((element == EPUB_XHTML_ELEMENT_BR) || (element == EPUB_XHTML_ELEMENT_HR)) {
			/*Just empty line, nothing else*/
			g_signal_emit_by_name(G_OBJECT(self), "text-chunk-ready", "\n", 1);
		}
		
		lark_document_loader_fb2_set_element_id_at_level(self, self->xmllevel-1, EPUB_XHTML_ELEMENT_UNKNOWN);
		self->xmllevel--;
	}
}

static void lark_document_loader_epub_xhtml_on_characters(gpointer data, const xmlChar *ch, gint len)
{
	LarkDocumentLoader *self = (LarkDocumentLoader *)data;
	gchar chars[len + 1];
	guint stelement, prelement;
	
	g_return_if_fail(self != NULL);
	
	strncpy(chars, (const gchar *)ch, len);
	chars[len] = '\0';
	
	stelement = lark_document_loader_fb2_get_element_id_at_level(self, self->xmllevel-1);
	
	if (lark_document_loader_fb2_get_element_id_at_level(self, 1) == EPUB_XHTML_ELEMENT_BODY) {
		/*Text section*/
		if ((stelement == EPUB_XHTML_ELEMENT_PARAGRAPH) || (stelement == EPUB_XHTML_ELEMENT_DIV) || (stelement == EPUB_XHTML_ELEMENT_ARTICLE) || (stelement == EPUB_XHTML_ELEMENT_BLOCKQUOTE)) {
			/*Text paragraph*/
			g_signal_emit_by_name(G_OBJECT(self), "text-chunk-ready", chars, len);
		} else if ((stelement == EPUB_XHTML_ELEMENT_H1) || (stelement == EPUB_XHTML_ELEMENT_H2) || (stelement == EPUB_XHTML_ELEMENT_H3) || (stelement == EPUB_XHTML_ELEMENT_H4) || (stelement == EPUB_XHTML_ELEMENT_H5) || (stelement == EPUB_XHTML_ELEMENT_H6)){
			/*Header*/
			g_signal_emit_by_name(G_OBJECT(self), "text-chunk-ready", chars, len);
		} else if (stelement == EPUB_XHTML_ELEMENT_STRONG) {
			prelement = lark_document_loader_fb2_get_element_id_at_level(self, self->xmllevel-2);
			if ((prelement == EPUB_XHTML_ELEMENT_PARAGRAPH) || (prelement == EPUB_XHTML_ELEMENT_DIV) || (prelement == EPUB_XHTML_ELEMENT_ARTICLE) || (prelement == EPUB_XHTML_ELEMENT_BLOCKQUOTE)) {
				/*Bold text*/
				g_signal_emit_by_name(G_OBJECT(self), "text-chunk-ready", chars, len);
			}
		}  else if (stelement == EPUB_XHTML_ELEMENT_ITALIC) {
			prelement = lark_document_loader_fb2_get_element_id_at_level(self, self->xmllevel-2);
			if ((prelement == EPUB_XHTML_ELEMENT_PARAGRAPH) || (prelement == EPUB_XHTML_ELEMENT_DIV) || (prelement == EPUB_XHTML_ELEMENT_ARTICLE) || (prelement == EPUB_XHTML_ELEMENT_BLOCKQUOTE)) {
				/*Italic text*/
				g_signal_emit_by_name(G_OBJECT(self), "text-chunk-ready", chars, len);
			}
		} else if (stelement == EPUB_XHTML_ELEMENT_BOLD) {
			prelement = lark_document_loader_fb2_get_element_id_at_level(self, self->xmllevel-2);
			if ((prelement == EPUB_XHTML_ELEMENT_PARAGRAPH) || (prelement == EPUB_XHTML_ELEMENT_DIV) || (prelement == EPUB_XHTML_ELEMENT_ARTICLE) || (prelement == EPUB_XHTML_ELEMENT_BLOCKQUOTE)) {
				/*Bold text*/
				g_signal_emit_by_name(G_OBJECT(self), "text-chunk-ready", chars, len);
			}
		} else if (stelement == EPUB_XHTML_ELEMENT_UNDERLINE) {
			prelement = lark_document_loader_fb2_get_element_id_at_level(self, self->xmllevel-2);
			if ((prelement == EPUB_XHTML_ELEMENT_PARAGRAPH) || (prelement == EPUB_XHTML_ELEMENT_DIV) || (prelement == EPUB_XHTML_ELEMENT_ARTICLE) || (prelement == EPUB_XHTML_ELEMENT_BLOCKQUOTE)) {
				/*Underline text*/
				g_signal_emit_by_name(G_OBJECT(self), "text-chunk-ready", chars, len);
			}
		} else if (stelement == EPUB_XHTML_ELEMENT_STRIKETHROUGH) {
			prelement = lark_document_loader_fb2_get_element_id_at_level(self, self->xmllevel-2);
			if ((prelement == EPUB_XHTML_ELEMENT_PARAGRAPH) || (prelement == EPUB_XHTML_ELEMENT_DIV) || (prelement == EPUB_XHTML_ELEMENT_ARTICLE) || (prelement == EPUB_XHTML_ELEMENT_BLOCKQUOTE)) {
				/*Strikethrough text*/
				g_signal_emit_by_name(G_OBJECT(self), "text-chunk-ready", chars, len);
			}
		} else if (stelement == EPUB_XHTML_ELEMENT_EM) {
			prelement = lark_document_loader_fb2_get_element_id_at_level(self, self->xmllevel-2);
			if ((prelement == EPUB_XHTML_ELEMENT_PARAGRAPH) || (prelement == EPUB_XHTML_ELEMENT_DIV) || (prelement == EPUB_XHTML_ELEMENT_ARTICLE) || (prelement == EPUB_XHTML_ELEMENT_BLOCKQUOTE)) {
				/*Italic text*/
				g_signal_emit_by_name(G_OBJECT(self), "text-chunk-ready", chars, len);
			}
		} else if (stelement == EPUB_XHTML_ELEMENT_SPAN) {
			prelement = lark_document_loader_fb2_get_element_id_at_level(self, self->xmllevel-2);
			if ((prelement == EPUB_XHTML_ELEMENT_PARAGRAPH) || (prelement == EPUB_XHTML_ELEMENT_DIV) || (prelement == EPUB_XHTML_ELEMENT_ARTICLE) || (prelement == EPUB_XHTML_ELEMENT_BLOCKQUOTE)) {
				/*Spanned text*/
				g_signal_emit_by_name(G_OBJECT(self), "text-chunk-ready", chars, len);
			}
		} else if (stelement == EPUB_XHTML_ELEMENT_SUP) {
			prelement = lark_document_loader_fb2_get_element_id_at_level(self, self->xmllevel-2);
			if ((prelement == EPUB_XHTML_ELEMENT_PARAGRAPH) || (prelement == EPUB_XHTML_ELEMENT_DIV) || (prelement == EPUB_XHTML_ELEMENT_ARTICLE) || (prelement == EPUB_XHTML_ELEMENT_BLOCKQUOTE)) {
				/*Superscript text*/
				g_signal_emit_by_name(G_OBJECT(self), "text-chunk-ready", chars, len);
			}
		} else if (stelement == EPUB_XHTML_ELEMENT_SUB) {
			prelement = lark_document_loader_fb2_get_element_id_at_level(self, self->xmllevel-2);
			if ((prelement == EPUB_XHTML_ELEMENT_PARAGRAPH) || (prelement == EPUB_XHTML_ELEMENT_DIV) || (prelement == EPUB_XHTML_ELEMENT_ARTICLE) || (prelement == EPUB_XHTML_ELEMENT_BLOCKQUOTE)) {
				/*Subscript text*/
				g_signal_emit_by_name(G_OBJECT(self), "text-chunk-ready", chars, len);
			}
		} else if (stelement == EPUB_XHTML_ELEMENT_A) {
			/*Link name*/
			g_signal_emit_by_name(G_OBJECT(self), "text-chunk-ready", chars, len);
		} else if (stelement == EPUB_XHTML_ELEMENT_TD) {
			prelement = lark_document_loader_fb2_get_element_id_at_level(self, self->xmllevel-2);
			if ((prelement == EPUB_XHTML_ELEMENT_TH) || (prelement == EPUB_XHTML_ELEMENT_TR)) {
				/*Table cell value*/
				g_signal_emit_by_name(G_OBJECT(self), "text-chunk-ready", chars, len);
			}
		} else if (stelement == EPUB_XHTML_ELEMENT_LI) {
			prelement = lark_document_loader_fb2_get_element_id_at_level(self, self->xmllevel-2);
			if ((stelement == EPUB_XHTML_ELEMENT_OL) || (stelement == EPUB_XHTML_ELEMENT_UL)) {
				/*List item*/
				g_signal_emit_by_name(G_OBJECT(self), "text-chunk-ready", chars, len);
			}
		}
	}
}

static void lark_document_loader_get_property (GObject *object, guint prop_id, GValue *value, GParamSpec *pspec)
{
	LarkDocumentLoader *self = (LarkDocumentLoader *)object;
	
	switch (prop_id) {
		case PROP_PATH:
			g_value_set_string(value, self->path);
			break;
		case PROP_FORMAT:
			g_value_set_uint(value, self->format);
			break;
		/*case PROP_TEXT:
			g_value_set_string(value, self->text);
			break;*/
		case PROP_TITLE:
			g_value_set_string(value, self->title);
			break;
		case PROP_AUTHOR:
			g_value_set_string(value, self->author);
			break;
		case PROP_COMMENT:
			g_value_set_string(value, self->comment);
			break;
		default:
			break;
	}
}

static void lark_document_loader_set_property (GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec)
{
	LarkDocumentLoader *self = (LarkDocumentLoader *)object;
	
	switch (prop_id) {
		case PROP_PATH:
			if (self->path != NULL) {
				g_free(self->path);
			}
			self->path = g_value_dup_string(value);
			break;
		case PROP_FORMAT:
			self->format = g_value_get_uint(value);
			break;
		/*case PROP_TEXT:
			if (self->text != NULL) {
				g_free(self->text);
			}
			self->text = g_value_dup_string(value);
			break;*/
		case PROP_TITLE:
			if (self->title != NULL) {
				g_free(self->title);
			}
			self->title = g_value_dup_string(value);
			break;
		case PROP_AUTHOR:
			if (self->author != NULL) {
				g_free(self->author);
			}
			self->author = g_value_dup_string(value);
			break;
		case PROP_COMMENT:
			if (self->comment != NULL) {
				g_free(self->comment);
			}
			self->comment = g_value_dup_string(value);
			break;
		default:
			break;
	}
}

static void lark_document_loader_constructed(GObject *object)
{
	LarkDocumentLoader *self __attribute__((unused)) = (LarkDocumentLoader *)object;
	
	G_OBJECT_CLASS(lark_document_loader_parent_class)->constructed(object);
}

static void lark_document_loader_finalize(GObject *object)
{
	LarkDocumentLoader *self = (LarkDocumentLoader *)object;
	
	if (self->path != NULL) {
		g_free(self->path);
	}
	
	/*if (self->text != NULL) {
		g_free(self->text);
	}*/
	
	if (self->title != NULL) {
		g_free(self->title);
	}
	
	if (self->author != NULL) {
		g_free(self->author);
	}
	
	if (self->comment != NULL) {
		g_free(self->comment);
	}
	
	G_OBJECT_CLASS(lark_document_loader_parent_class)->finalize(object);
}

static void lark_document_loader_class_init(LarkDocumentLoaderClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS(klass);
	
	object_class->get_property = lark_document_loader_get_property;
	object_class->set_property = lark_document_loader_set_property;
	
	object_class->constructed = lark_document_loader_constructed;
	object_class->finalize = lark_document_loader_finalize;
	
	properties[PROP_PATH] =
		g_param_spec_string ("path",
							"Document path",
							"Opened document path",
							NULL,
							G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE);
	
	properties[PROP_FORMAT] =
		g_param_spec_uint ("format",
							"Document format",
							"Opened document format identifier",
							LARK_DOCUMENT_LOADER_FORMAT_UNKNOWN,
							LARK_DOCUMENT_LOADER_FORMAT_FB2_ZIP,
							LARK_DOCUMENT_LOADER_FORMAT_UNKNOWN,
							G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE);
	
	/*properties[PROP_TEXT] =
		g_param_spec_string ("text",
							"Document text",
							"Opened document text",
							NULL,
							G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE);*/
	
	properties[PROP_TITLE] =
		g_param_spec_string ("title",
							"Document title",
							"Opened document title",
							NULL,
							G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE);
	
	properties[PROP_AUTHOR] =
		g_param_spec_string ("author",
							"Document author name",
							"Opened document author name",
							NULL,
							G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE);
	
	properties[PROP_COMMENT] =
		g_param_spec_string ("comment",
							"Document comment",
							"Opened document comment",
							NULL,
							G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE);
	
	g_object_class_install_properties (object_class, LAST_PROP, properties);
	
	signals[TEXT_CHUNK_READY] =
		g_signal_new("text-chunk-ready",
							G_TYPE_FROM_CLASS(klass),
							G_SIGNAL_RUN_LAST,
							0,
							NULL,
							NULL,
							g_cclosure_marshal_generic,
							G_TYPE_NONE,
							2,
							G_TYPE_STRING,
							G_TYPE_UINT);
}

static void lark_document_loader_init(LarkDocumentLoader *self)
{
	
}

gboolean lark_document_loader_open_text_file(LarkDocumentLoader *self, gchar *path, GError **error)
{
	GIOChannel *filestream;
	GIOStatus status;
	gchar buffer[1024];
	gsize bytesread;
	gchar *line;
	gsize length;
	uchardet_t chardet;
	gchar *charset;
	
	g_return_val_if_fail(self != NULL,  FALSE);
	g_return_val_if_fail(path != NULL,  FALSE);
	g_return_val_if_fail(error != NULL, FALSE);
	
	filestream = g_io_channel_new_file(path, "r", error);
	if (filestream == NULL) {
		return FALSE;
	}
	/*Disable reencoding functionality*/
	status = g_io_channel_set_encoding(filestream, NULL, error);
	if (status != G_IO_STATUS_NORMAL) {
		g_io_channel_shutdown(filestream, TRUE, NULL);
		return FALSE;
	}
	/*Read buffer for encoding detection*/
	memset(buffer, 0, sizeof(buffer));
	status = g_io_channel_read_chars(filestream, (gchar *)buffer, sizeof(buffer), &bytesread, error);
	if (status != G_IO_STATUS_NORMAL) {
		g_io_channel_shutdown(filestream, TRUE, NULL);
		return FALSE;
	}
	/*Try to detect charset*/
	if (bytesread > 0) {
		/*New uchardet lib instance*/
		chardet = uchardet_new();
		/*Send buffer for charset detection*/
		if (uchardet_handle_data(chardet, buffer, bytesread)) {
			g_io_channel_shutdown(filestream, TRUE, NULL);
			uchardet_delete(chardet);
			return FALSE;
		}
		/*Buffer is sent, time to detect encoding*/
		uchardet_data_end(chardet);
		/*Set charset for automatic reencoding*/
		charset = (gchar *)uchardet_get_charset(chardet);
		if ((charset != NULL) && (charset[0] != '\0')) {
			status = g_io_channel_set_encoding(filestream, charset, error);
			if (status != G_IO_STATUS_NORMAL) {
				g_io_channel_shutdown(filestream, TRUE, NULL);
				uchardet_delete(chardet);
				return FALSE;
			}
		}
		/*Free uchardet lib instance*/
		uchardet_delete(chardet);
	}
	/*Rewind stream*/
	status = g_io_channel_seek_position(filestream, 0, G_SEEK_SET, error);
	if (status != G_IO_STATUS_NORMAL) {
		g_io_channel_shutdown(filestream, TRUE, NULL);
		return FALSE;
	}
	/*Read file contents*/
	do {
		status = g_io_channel_read_line(filestream, &line, &length, NULL, error);
		if (status == G_IO_STATUS_NORMAL) {
			if (line != NULL) {
				g_signal_emit_by_name(G_OBJECT(self), "text-chunk-ready", line, length);
				g_free(line);
			}
		} else if (status == G_IO_STATUS_ERROR) {
			g_io_channel_shutdown(filestream, TRUE, NULL);
			return FALSE;
		}
	} while (status == G_IO_STATUS_NORMAL);
	/*Close IO stream*/
	status = g_io_channel_shutdown(filestream, TRUE, error);
	if (status != G_IO_STATUS_NORMAL) {
		return FALSE;
	}
	/*Update metadata*/
	self->path = g_strdup(path);
	self->format = LARK_DOCUMENT_LOADER_FORMAT_TXT;
	
	return TRUE;
}

gboolean lark_document_loader_open_pdf_file(LarkDocumentLoader *self, gchar *path, GError **error)
{
	GFile *pdffile;
	PopplerDocument *pdfdoc;
	gint pdfpages, p;
	PopplerPage *pdfpage;
	gchar *line;
	
	g_return_val_if_fail(self != NULL,  FALSE);
	g_return_val_if_fail(path != NULL,  FALSE);
	g_return_val_if_fail(error != NULL, FALSE);
	
	pdffile = g_file_new_for_path((const gchar *)path);
	pdfdoc = poppler_document_new_from_gfile(pdffile, NULL, NULL, error);
	if (pdfdoc == NULL) {
		return FALSE;
	}
	/*Read text page by page*/
	pdfpages = poppler_document_get_n_pages(pdfdoc);
	for (p = 0; p < pdfpages; p++) {
		pdfpage = poppler_document_get_page(pdfdoc, p);
		line = poppler_page_get_text(pdfpage);
		if (line != NULL) {
			g_signal_emit_by_name(G_OBJECT(self), "text-chunk-ready", line, strlen(line));
			g_free(line);
		}
	}
	/*Update metadata*/
	self->path = g_strdup(path);
	self->format = LARK_DOCUMENT_LOADER_FORMAT_PDF;
	self->title = poppler_document_get_title(pdfdoc);
	self->author = poppler_document_get_author(pdfdoc);
	self->comment = poppler_document_get_subject(pdfdoc);
	/*Free resources*/
	g_object_unref(pdfdoc);
	g_object_unref(pdffile);
	
	return TRUE;
}

gboolean lark_document_loader_open_epub_file(LarkDocumentLoader *self, gchar *path, GError **error)
{
	GsfInput *src, *arcfile;
	GsfInfile *zipfile;
	guint8 data[4096];
	gsf_off_t datasize;
	gsize readsize;
	guint xmlelement, filenameseq;
	xmlSAXHandler saxhandler;
	xmlSAXHandler htmlsaxhandler;
	xmlParserCtxtPtr xmlcontext;
	htmlParserCtxtPtr htmlcontext;
	GSList *iter;
	const gchar *contfile[] = {"META-INF", "container.xml", NULL};
	   	
	g_return_val_if_fail(self != NULL,  FALSE);
	g_return_val_if_fail(path != NULL,  FALSE);
	g_return_val_if_fail(error != NULL, FALSE);
	
	memset(data, 0, sizeof(data));
	
	/*Open file*/
	src = gsf_input_stdio_new(path, error);
	if (src == NULL) {
		return FALSE;
	}
	/*Parse file*/
	zipfile = gsf_infile_zip_new(src, error);
	if (zipfile == NULL) {
		g_object_unref(src);
		return FALSE;
	}
	g_object_unref(src);
	/*Test MIME type*/
	arcfile = gsf_infile_child_by_name(zipfile, "mimetype");
	if (arcfile == NULL) {
		*error = g_error_new(LARK_DOCUMENT_LOADER_ERROR, 6, _("Unable to get MIME type information."));
		g_object_unref(zipfile);
		return FALSE;
	}
	datasize = gsf_input_size(GSF_INPUT(arcfile));
	if ((datasize >= sizeof(data)) || (datasize == 0)) {
		*error = g_error_new(LARK_DOCUMENT_LOADER_ERROR, 6, _("Unable to get MIME type information."));
		g_object_unref(zipfile);
		return FALSE;
	}
	gsf_input_read(GSF_INPUT(arcfile), datasize, data);
	if (!g_str_equal(data, "application/epub+zip")) {
		*error = g_error_new(LARK_DOCUMENT_LOADER_ERROR, 7, _("File is of unknown MIME type: %s."), data);
		g_object_unref(zipfile);
		return FALSE;
	}
	g_object_unref(arcfile);
	/*Get name of OPF file*/
	self->epubopfpath = NULL;
	arcfile = gsf_infile_child_by_aname(zipfile, contfile);
	if (arcfile == NULL) {
		*error = g_error_new(LARK_DOCUMENT_LOADER_ERROR, 8, _("Unable to read metadata file."));
		g_object_unref(zipfile);
		return FALSE;
	}
	datasize = gsf_input_size(GSF_INPUT(arcfile));
	if (datasize == 0) {
		*error = g_error_new(LARK_DOCUMENT_LOADER_ERROR, 8, _("Unable to read metadata file."));
		g_object_unref(zipfile);
		return FALSE;
	}
	/*Use parser*/
	memset(&saxhandler, 0, sizeof(xmlSAXHandler));
	saxhandler.initialized = XML_SAX2_MAGIC;
	saxhandler.startElementNs = lark_document_loader_epub_container_on_start_element;
	xmlcontext = xmlCreatePushParserCtxt(&saxhandler, self, NULL, 0, NULL);
	while (datasize > 0) {
		if (datasize >= sizeof(data)) {
			readsize = sizeof(data);
		} else {
			readsize = datasize;
		}
		if (gsf_input_read(GSF_INPUT(arcfile), readsize, data) == NULL) {
			*error = g_error_new(LARK_DOCUMENT_LOADER_ERROR, 2, _("Unable to open selected file."));
			/*Close parser*/
			xmlFreeParserCtxt(xmlcontext);
			xmlCleanupParser();
			/*Free state variables*/
			g_object_unref(arcfile);
			g_object_unref(zipfile);
			return FALSE;
		}
		if (xmlParseChunk(xmlcontext, (const gchar *)data, readsize, 0) != 0) {
			*error = g_error_new(LARK_DOCUMENT_LOADER_ERROR, 4, _("Unable to parse file chunk."));
			/*Close parser*/
			xmlFreeParserCtxt(xmlcontext);
			xmlCleanupParser();
			/*Free state variables*/
			g_object_unref(arcfile);
			g_object_unref(zipfile);
			return FALSE;
		}
		datasize -= readsize;
	}
    /*Close parser*/
	xmlParseChunk(xmlcontext, NULL, 0, 1);
	xmlFreeParserCtxt(xmlcontext);
	xmlCleanupParser();
	g_object_unref(arcfile);
	/*Check if pages are available*/
	if (self->epubopfpath == NULL) {
		*error = g_error_new(LARK_DOCUMENT_LOADER_ERROR, 9, _("Unable to get list of OPF files."));
		/*Free state variables*/
		g_object_unref(zipfile);
		return FALSE;
	}
	/*Get table of (X)HTML files to parse*/
	arcfile = gsf_infile_child_by_aname(zipfile, (const gchar **)self->epubopfpath);
	if (arcfile == NULL) {
		*error = g_error_new(LARK_DOCUMENT_LOADER_ERROR, 8, _("Unable to read metadata file."));
		g_strfreev(self->epubopfpath);
		g_object_unref(zipfile);
		return FALSE;
	}
	datasize = gsf_input_size(GSF_INPUT(arcfile));
	if (datasize == 0) {
		*error = g_error_new(LARK_DOCUMENT_LOADER_ERROR, 8, _("Unable to read metadata file."));
		g_strfreev(self->epubopfpath);
		g_object_unref(zipfile);
		return FALSE;
	}
	/*Initialize state variables*/
	self->xmllevel = 0;
	self->xmlelements = g_array_new(FALSE, TRUE, sizeof(guint));
	self->xmlauthor = g_string_new(NULL);
	self->xmlcomment = g_string_new(NULL);
	xmlelement = EPUB_OPF_ELEMENT_UNKNOWN;
	g_array_append_val(self->xmlelements, xmlelement);
	self->epubitems = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, g_free);
	self->epubspine = NULL;
	/*Use parser*/
	memset(&saxhandler, 0, sizeof(xmlSAXHandler));
	saxhandler.initialized = XML_SAX2_MAGIC;
	saxhandler.startElementNs = lark_document_loader_epub_opf_on_start_element;
	saxhandler.endElementNs = lark_document_loader_epub_opf_on_end_element;
	saxhandler.characters = lark_document_loader_epub_opf_on_characters;
	xmlcontext = xmlCreatePushParserCtxt(&saxhandler, self, NULL, 0, NULL);
	while (datasize > 0) {
		if (datasize >= sizeof(data)) {
			readsize = sizeof(data);
		} else {
			readsize = datasize;
		}
		if (gsf_input_read(GSF_INPUT(arcfile), readsize, data) == NULL) {
			*error = g_error_new(LARK_DOCUMENT_LOADER_ERROR, 2, _("Unable to open selected file."));
			/*Close parser*/
			xmlFreeParserCtxt(xmlcontext);
			xmlCleanupParser();
			/*Free state variables*/
			self->xmllevel = 0;
			g_array_free(self->xmlelements, TRUE);
			g_string_free(self->xmlauthor, TRUE);
			g_string_free(self->xmlcomment, TRUE);
			g_hash_table_destroy(self->epubitems);
			g_slist_free(self->epubspine);
			g_strfreev(self->epubopfpath);
			g_object_unref(arcfile);
			g_object_unref(zipfile);
			return FALSE;
		}
		if (xmlParseChunk(xmlcontext, (const gchar *)data, readsize, 0) != 0) {
			*error = g_error_new(LARK_DOCUMENT_LOADER_ERROR, 4, _("Unable to parse file chunk."));
			/*Close parser*/
			xmlFreeParserCtxt(xmlcontext);
			xmlCleanupParser();
			/*Free state variables*/
			self->xmllevel = 0;
			g_array_free(self->xmlelements, TRUE);
			g_string_free(self->xmlauthor, TRUE);
			g_string_free(self->xmlcomment, TRUE);
			g_hash_table_destroy(self->epubitems);
			g_slist_free(self->epubspine);
			g_strfreev(self->epubopfpath);
			g_object_unref(arcfile);
			g_object_unref(zipfile);
			return FALSE;
		}
		datasize -= readsize;
	}
    /*Close parser*/
	self->xmllevel = 0;
	g_array_free(self->xmlelements, TRUE);
	xmlParseChunk(xmlcontext, NULL, 0, 1);
	xmlFreeParserCtxt(xmlcontext);
	xmlCleanupParser();
	g_object_unref(arcfile);
	/*Check if pages are available*/
	if (self->epubspine == NULL) {
		*error = g_error_new(LARK_DOCUMENT_LOADER_ERROR, 9, _("Unable to get list of chapters."));
		/*Free state variables*/
		g_string_free(self->xmlauthor, TRUE);
		g_string_free(self->xmlcomment, TRUE);
		g_hash_table_destroy(self->epubitems);
		g_strfreev(self->epubopfpath);
		g_object_unref(zipfile);
		return FALSE;
	}
	/*Parse (X)HTML pages one by one*/
	filenameseq = g_strv_length(self->epubopfpath) - 1;
	self->epubspine = g_slist_reverse(self->epubspine);
	for (iter = self->epubspine; iter != NULL; iter = iter->next) {
		/*Replace filename*/
		g_free(self->epubopfpath[filenameseq]);
		self->epubopfpath[filenameseq] = g_strdup((gchar *)iter->data);
		/*Open and parse file*/
		arcfile = gsf_infile_child_by_aname(zipfile, (const gchar **)self->epubopfpath);
		if (arcfile == NULL) {
			continue;
		}
		datasize = gsf_input_size(GSF_INPUT(arcfile));
		if (datasize == 0) {
			g_object_unref(arcfile);
			continue;
		}
		self->xmllevel = 0;
		self->xmlelements = g_array_new(FALSE, TRUE, sizeof(guint));
		memset(&htmlsaxhandler, 0, sizeof(htmlSAXHandler));
		htmlsaxhandler.initialized = XML_SAX2_MAGIC;
		htmlsaxhandler.startElement = lark_document_loader_epub_xhtml_on_start_element;
		htmlsaxhandler.endElement = lark_document_loader_epub_xhtml_on_end_element;
		htmlsaxhandler.characters = lark_document_loader_epub_xhtml_on_characters;
		htmlcontext = htmlCreatePushParserCtxt(&htmlsaxhandler, self, NULL, 0, NULL, XML_CHAR_ENCODING_UTF8);
		htmlCtxtUseOptions(htmlcontext, HTML_PARSE_NOBLANKS | HTML_PARSE_NOERROR | HTML_PARSE_NOWARNING | HTML_PARSE_NONET);
		while (datasize > 0) {
			if (datasize >= sizeof(data)) {
				readsize = sizeof(data);
			} else {
				readsize = datasize;
			}
			if (gsf_input_read(GSF_INPUT(arcfile), readsize, data) == NULL) {
				break;
			}
			if (htmlParseChunk(htmlcontext, (const gchar *)data, readsize, 0) != 0) {
				break;
			}
			datasize -= readsize;
		}
		htmlParseChunk(htmlcontext, NULL, 0, 1);
		htmlFreeParserCtxt(htmlcontext);
		xmlCleanupParser();
		g_object_unref(arcfile);
		g_array_free(self->xmlelements, TRUE);
	}
	/*Update metadata*/
    self->path = g_strdup(path);
    self->format = LARK_DOCUMENT_LOADER_FORMAT_EPUB;
	self->author = g_string_free(self->xmlauthor, FALSE);
	self->comment = g_string_free(self->xmlcomment, FALSE);
	/*Free state variables*/
	g_slist_free(self->epubspine);
	g_hash_table_destroy(self->epubitems);
	g_strfreev(self->epubopfpath);
	g_object_unref(zipfile);
	
	return TRUE;
}

gboolean lark_document_loader_open_fb2_file(LarkDocumentLoader *self, gchar *path, GError **error)
{
	guint xmlelement;
	xmlSAXHandler saxhandler;
	xmlParserCtxtPtr xmlcontext;
	GIOChannel *filestream;
	GIOStatus status;
	gchar *line;
	gsize length;
		
	g_return_val_if_fail(self != NULL,  FALSE);
	g_return_val_if_fail(path != NULL,  FALSE);
	g_return_val_if_fail(error != NULL, FALSE);
	
	filestream = g_io_channel_new_file(path, "r", error);
	if (filestream == NULL) {
		return FALSE;
	}
	/*Disable reencoding functionality*/
	status = g_io_channel_set_encoding(filestream, NULL, error);
	if (status != G_IO_STATUS_NORMAL) {
		g_io_channel_shutdown(filestream, TRUE, NULL);
		return FALSE;
	}
	/*Initialize state variables*/
	self->xmllevel = 0;
	self->xmlelements = g_array_new(FALSE, TRUE, sizeof(guint));
	self->xmlauthor = g_string_new(NULL);
	self->xmlcomment = g_string_new(NULL);
	xmlelement = FB2_ELEMENT_UNKNOWN;
	g_array_append_val(self->xmlelements, xmlelement);
	/*SAX parser callbacks*/
	memset(&saxhandler, 0, sizeof(xmlSAXHandler));
	saxhandler.initialized = XML_SAX2_MAGIC;
	saxhandler.startElement = lark_document_loader_fb2_on_start_element;
	saxhandler.endElement = lark_document_loader_fb2_on_end_element;
	saxhandler.characters = lark_document_loader_fb2_on_characters;
	/*SAX context*/
	xmlcontext = xmlCreatePushParserCtxt(&saxhandler, self, NULL, 0, NULL);
	/*Feed data*/
	do {
		status = g_io_channel_read_line(filestream, &line, &length, NULL, error);
		if (status == G_IO_STATUS_NORMAL) {
			if (line != NULL) {
				if (xmlParseChunk(xmlcontext, (const gchar *)line, length, 0) != 0) {
					*error = g_error_new(LARK_DOCUMENT_LOADER_ERROR, 4, _("Unable to parse file chunk."));
					/*Close parser*/
					xmlFreeParserCtxt(xmlcontext);
					xmlCleanupParser();
					/*Free state variables*/
					self->xmllevel = 0;
					g_array_free(self->xmlelements, TRUE);
					g_string_free(self->xmlauthor, TRUE);
					g_string_free(self->xmlcomment, TRUE);
					/*Close stream*/
					g_io_channel_shutdown(filestream, TRUE, NULL);
					/*Free last line*/
					g_free(line);
					return FALSE;
				}
				g_free(line);
			}
		} else if (status == G_IO_STATUS_ERROR) {
			/*Close parser*/
			xmlFreeParserCtxt(xmlcontext);
			xmlCleanupParser();
			/*Free state variables*/
			self->xmllevel = 0;
			g_array_free(self->xmlelements, TRUE);
			g_string_free(self->xmlauthor, TRUE);
			g_string_free(self->xmlcomment, TRUE);
			/*Close stream*/
			g_io_channel_shutdown(filestream, TRUE, NULL);
			return FALSE;
		}
	} while (status == G_IO_STATUS_NORMAL);
	/*Close IO stream*/
	status = g_io_channel_shutdown(filestream, TRUE, error);
	if (status != G_IO_STATUS_NORMAL) {
		return FALSE;
	}
	/*Update metadata*/
	self->path = g_strdup(path);
	self->format = LARK_DOCUMENT_LOADER_FORMAT_FB2;
	self->author = g_string_free(self->xmlauthor, FALSE);
	self->comment = g_string_free(self->xmlcomment, FALSE);
	/*Close parser*/
	xmlParseChunk(xmlcontext, NULL, 0, 1);
	xmlFreeParserCtxt(xmlcontext);
	xmlCleanupParser();
	/*Free state variables*/
	self->xmllevel = 0;
	g_array_free(self->xmlelements, TRUE);
	
	return TRUE;
}

gboolean lark_document_loader_open_fb2_zip_file(LarkDocumentLoader *self, gchar *path, GError **error)
{
	GsfInput *src, *arcfile;
	GsfInfile *zipfile;
	gchar *fnstart, *fnend, *fname;
	guint8 data[4096];
	gsf_off_t datasize;
	gsize readsize;
	guint xmlelement;
	xmlSAXHandler saxhandler;
	xmlParserCtxtPtr xmlcontext;
	
	g_return_val_if_fail(self != NULL,  FALSE);
	g_return_val_if_fail(path != NULL,  FALSE);
	g_return_val_if_fail(error != NULL, FALSE);
	
	memset(data, 0, sizeof(data));
	
	/*Open file*/
	src = gsf_input_stdio_new(path, error);
	if (src == NULL) {
		return FALSE;
	}
	/*Parse file*/
	zipfile = gsf_infile_zip_new(src, error);
	if (zipfile == NULL) {
		g_object_unref(src);
		return FALSE;
	}
	g_object_unref(src);
	/*Get FB2 filename*/
	fnstart = strrchr(path, G_DIR_SEPARATOR);
	if (fnstart == NULL) {
		fnstart = path;
	} else {
		fnstart = fnstart + 1;
	}
	fnend = strrchr(path, '.');
	if (fnend != NULL) {
		fname = g_strndup(fnstart, fnend - fnstart);
	} else {
		fname = g_strndup(fnstart, strlen(fnstart));
	}
	/*Unpack file*/
	arcfile = gsf_infile_child_by_name(zipfile, (const gchar *)fname);
	if (arcfile == NULL) {
		*error = g_error_new(LARK_DOCUMENT_LOADER_ERROR, 8, _("Unable to read selected file."));
		g_free(fname);
		g_object_unref(zipfile);
		return FALSE;
	}
	g_free(fname);
	datasize = gsf_input_size(GSF_INPUT(arcfile));
	if (datasize == 0) {
		*error = g_error_new(LARK_DOCUMENT_LOADER_ERROR, 8, _("Unable to read selected file."));
		g_object_unref(zipfile);
		return FALSE;
	}
	/*Initialize state variables*/
	self->xmllevel = 0;
	self->xmlelements = g_array_new(FALSE, TRUE, sizeof(guint));
	self->xmlauthor = g_string_new(NULL);
	self->xmlcomment = g_string_new(NULL);
	xmlelement = FB2_ELEMENT_UNKNOWN;
	g_array_append_val(self->xmlelements, xmlelement);
	/*SAX parser callbacks*/
	memset(&saxhandler, 0, sizeof(xmlSAXHandler));
	saxhandler.initialized = XML_SAX2_MAGIC;
	saxhandler.startElement = lark_document_loader_fb2_on_start_element;
	saxhandler.endElement = lark_document_loader_fb2_on_end_element;
	saxhandler.characters = lark_document_loader_fb2_on_characters;
	/*SAX context*/
	xmlcontext = xmlCreatePushParserCtxt(&saxhandler, self, NULL, 0, NULL);
	while (datasize > 0) {
		if (datasize >= sizeof(data)) {
			readsize = sizeof(data);
		} else {
			readsize = datasize;
		}
		if (gsf_input_read(GSF_INPUT(arcfile), readsize, data) == NULL) {
			*error = g_error_new(LARK_DOCUMENT_LOADER_ERROR, 2, _("Unable to open selected file."));
			/*Close parser*/
			xmlFreeParserCtxt(xmlcontext);
			xmlCleanupParser();
			/*Free state variables*/
			self->xmllevel = 0;
			g_array_free(self->xmlelements, TRUE);
			g_string_free(self->xmlauthor, TRUE);
			g_string_free(self->xmlcomment, TRUE);
			g_object_unref(arcfile);
			g_object_unref(zipfile);
			return FALSE;
		}
		if (xmlParseChunk(xmlcontext, (const gchar *)data, readsize, 0) != 0) {
			*error = g_error_new(LARK_DOCUMENT_LOADER_ERROR, 4, _("Unable to parse file chunk."));
			/*Close parser*/
			xmlFreeParserCtxt(xmlcontext);
			xmlCleanupParser();
			/*Free state variables*/
			self->xmllevel = 0;
			g_array_free(self->xmlelements, TRUE);
			g_string_free(self->xmlauthor, TRUE);
			g_string_free(self->xmlcomment, TRUE);
			g_object_unref(arcfile);
			g_object_unref(zipfile);
			return FALSE;
		}
		datasize -= readsize;
	}
    /*Update metadata*/
	self->path = g_strdup(path);
	self->format = LARK_DOCUMENT_LOADER_FORMAT_FB2_ZIP;
	self->author = g_string_free(self->xmlauthor, FALSE);
	self->comment = g_string_free(self->xmlcomment, FALSE);
	/*Close parser*/
	xmlParseChunk(xmlcontext, (const gchar *)data, 0, 1);
	xmlFreeParserCtxt(xmlcontext);
	xmlCleanupParser();
	/*Free state variables*/
	self->xmllevel = 0;
	g_array_free(self->xmlelements, TRUE);
	g_object_unref(arcfile);
	g_object_unref(zipfile);
	
	return TRUE;
}

gboolean lark_document_loader_open(LarkDocumentLoader *self, gchar *path, GError **error)
{
	GStatBuf buf;
	gboolean res;
	gchar *lcpath;
	
	g_return_val_if_fail(path != NULL, FALSE);
	
	if ((g_stat(path, &buf) == 0) && (g_access(path, R_OK) == 0)) {
		if (S_ISREG(buf.st_mode)) {
			/*Free metadata*/
			if (self->title != NULL) {
				g_free(self->title);
				self->title = NULL;
			}
			if (self->author != NULL) {
				g_free(self->author);
				self->author = NULL;
			}
			if (self->comment != NULL) {
				g_free(self->comment);
				self->comment = NULL;
			}
			if (self->path != NULL) {
				g_free(self->path);
				self->path = NULL;
			}
			self->format = LARK_DOCUMENT_LOADER_FORMAT_UNKNOWN;
			/*Get lowercase path*/
			lcpath = g_utf8_strdown(path, -1);
			/*Parse supported files*/
			if (g_str_has_suffix(lcpath, ".txt")) {
				/*Raw text file*/
				res = lark_document_loader_open_text_file(self, path, error);
			} else if (g_str_has_suffix(lcpath, ".pdf")) {
				/*PDF file*/
				res = lark_document_loader_open_pdf_file(self, path, error);
			} else if (g_str_has_suffix(lcpath, ".epub")) {
				/*EPUB file*/
				res = lark_document_loader_open_epub_file(self, path, error);
			} else if (g_str_has_suffix(lcpath, ".fb2")) {
				/*FB2 xml file*/
				res = lark_document_loader_open_fb2_file(self, path, error);
			} else if (g_str_has_suffix(lcpath, ".fb2.zip")) {
				/*Compressed FB2 xml file*/
				res = lark_document_loader_open_fb2_zip_file(self, path, error);
			} else {
				if (error != NULL) {
					*error = g_error_new(LARK_DOCUMENT_LOADER_ERROR, 1, _("File format isn\'n supported."));
				}
				g_free(lcpath);
				return FALSE;
			}
			g_free(lcpath);
		} else {
			if (error != NULL) {
				*error = g_error_new(LARK_DOCUMENT_LOADER_ERROR, 5, _("File is of unknown type."));
			}
			return FALSE;
		}
	} else {
		if (error != NULL) {
			*error = g_error_new(LARK_DOCUMENT_LOADER_ERROR, 0, _("Unable to read file."));
		}
		return FALSE;
	}
	
	return res;
}

gchar *lark_document_loader_get_filename(LarkDocumentLoader *self)
{
	gchar *fnstart, *lcstart, *lcend;
	gsize fnlen;
	
	g_return_val_if_fail(self != NULL, NULL);
	g_return_val_if_fail(self->path != NULL, NULL);
	
	fnstart = strrchr(self->path, G_DIR_SEPARATOR);
	
	if (fnstart == NULL) {
		fnstart = self->path;
	} else {
		fnstart = fnstart + 1;
	}
	
	fnlen = strlen(fnstart);
	
	lcstart = g_utf8_strdown((const gchar *)fnstart, fnlen);
	
	switch (self->format) {
		case LARK_DOCUMENT_LOADER_FORMAT_TXT:
			lcend = g_strrstr((const gchar *)lcstart, ".txt");
			break;
		case LARK_DOCUMENT_LOADER_FORMAT_PDF:
			lcend = g_strrstr((const gchar *)lcstart, ".pdf");
			break;
		case LARK_DOCUMENT_LOADER_FORMAT_EPUB:
			lcend = g_strrstr((const gchar *)lcstart, ".epub");
			break;
		case LARK_DOCUMENT_LOADER_FORMAT_FB2:
			lcend = g_strrstr((const gchar *)lcstart, ".fb2");
			break;
		case LARK_DOCUMENT_LOADER_FORMAT_FB2_ZIP:
			lcend = g_strrstr((const gchar *)lcstart, ".fb2.zip");
			break;
		default:
			lcend = strrchr(lcstart, '.');
			break;
	}
	
	if (lcend != NULL) {
		fnlen = lcend - lcstart;
	}
	
	g_free(lcstart);
	
	return g_strndup(fnstart, fnlen);
}

gchar *lark_document_loader_get_title(LarkDocumentLoader *self)
{
	g_return_val_if_fail(self != NULL, NULL);
	
	return self->title;
}

gchar *lark_document_loader_get_author(LarkDocumentLoader *self)
{
	g_return_val_if_fail(self != NULL, NULL);
	
	return self->author;
}

gchar *lark_document_loader_get_comment(LarkDocumentLoader *self)
{
	g_return_val_if_fail(self != NULL, NULL);
	
	return self->comment;
}

LarkDocumentLoader *lark_document_loader_new(void)
{
	return g_object_new(lark_document_loader_get_type(), NULL);
}
